"""
comparison of ranked lists

"""


# from rbo import RankingSimilarity


# def compare_rbo(df1, df2, p):

#     return RankingSimilarity(df1.index, df2.index).rbo(p=p)


# def compare_gwets_ac1(df1, df2):

#     pass
